from django_scss.utils import compile_scss, logger
from django_scss.settings import SCSS_DEVMODE_WATCH_DIRS, SCSS_OUTPUT_DIR, SCSS_DEVMODE_EXCLUDE
from .utils import STATIC_ROOT
import os
import re
import sys
import time
import threading


WATCHED_FILES = {}
SCSS_IMPORT_RE = re.compile(r"""@import\s+['"](.+?\.scss)['"]\s*;""")


def daemon():

    while True:
        to_be_compiled = set()
        for watched_dir in SCSS_DEVMODE_WATCH_DIRS:
            for root, dirs, files in os.walk(watched_dir):
                for filename in filter(lambda f: f.endswith(".scss"), files):
                    filename = os.path.join(root, filename)
                    f = os.path.relpath(filename, STATIC_ROOT)
                    if f in SCSS_DEVMODE_EXCLUDE:
                        continue
                    mtime = os.path.getmtime(filename)

                    if f not in WATCHED_FILES:
                        WATCHED_FILES[f] = [None, set()]

                    if WATCHED_FILES[f][0] != mtime:
                        WATCHED_FILES[f][0] = mtime
                        # Look for @import statements to update dependecies
                        for line in open(filename):
                            for imported in SCSS_IMPORT_RE.findall(line):
                                imported = os.path.relpath(os.path.join(os.path.dirname(filename), imported), STATIC_ROOT)
                                if imported not in WATCHED_FILES:
                                    WATCHED_FILES[imported] = [None, set([f])]
                                else:
                                    WATCHED_FILES[imported][1].add(f)

                        to_be_compiled.add(f)
                        for importer in WATCHED_FILES[f][1]:
                            to_be_compiled.add(importer)

        for scss_path in to_be_compiled:
            full_path = os.path.join(STATIC_ROOT, scss_path)
            base_filename = os.path.split(scss_path)[-1][:-5]
            output_directory = os.path.join(STATIC_ROOT, SCSS_OUTPUT_DIR, os.path.dirname(scss_path))
            output_path = os.path.join(output_directory, "%s.css" % base_filename)
            if isinstance(full_path, unicode):
                filesystem_encoding = sys.getfilesystemencoding() or sys.getdefaultencoding()
                full_path = full_path.encode(filesystem_encoding)

            compile_scss(full_path, output_path, scss_path)
            logger.debug("Compiled: %s" % scss_path)


        time.sleep(1)


def start_daemon():
    thread = threading.Thread(target=daemon)
    thread.daemon = True
    thread.start()




from django.conf import settings


SCSS_EXECUTABLE = getattr(settings, "SCSS_EXECUTABLE", "sass")
SCSS_USE_CACHE = getattr(settings, "SCSS_USE_CACHE", True)
SCSS_CACHE_TIMEOUT = getattr(settings, "SCSS_CACHE_TIMEOUT", 60 * 60 * 24 * 30) # 30 days
SCSS_MTIME_DELAY = getattr(settings, "SCSS_MTIME_DELAY", 10) # 10 seconds
SCSS_ROOT = getattr(settings, "SCSS_ROOT", getattr(settings, "STATIC_ROOT", getattr(settings, "MEDIA_ROOT")))
SCSS_OUTPUT_DIR = getattr(settings, "SCSS_OUTPUT_DIR", "SCSS_CACHE")
SCSS_DEVMODE = getattr(settings, "SCSS_DEVMODE", False)
SCSS_DEVMODE_WATCH_DIRS = getattr(settings, "SCSS_DEVMODE_WATCH_DIRS", [settings.STATIC_ROOT])
SCSS_DEVMODE_EXCLUDE = getattr(settings, "SCSS_DEVMODE_EXCLUDE", ())
SCSS_USE_COMPASS = getattr(settings, "SCSS_USE_COMPASS", False)



from django_scss.storage import SCSSFileStorage
from django.contrib.staticfiles.finders import BaseStorageFinder


class SCSSFinder(BaseStorageFinder):
    """
    A staticfiles finder that looks in SCSS_ROOT
    for compiled files, to be used during development
    with staticfiles development file server or during
    deployment.
    """
    storage = SCSSFileStorage

    def list(self, ignore_patterns):
        return []



from django.core.files.storage import FileSystemStorage
from django_scss.settings import SCSS_ROOT


class SCSSFileStorage(FileSystemStorage):
    """
    Standard file system storage for files handled by django-scss.

    The default for ``location`` is ``SCSS_ROOT``
    """
    def __init__(self, location=None, base_url=None, *args, **kwargs):
        if location is None:
            location = SCSS_ROOT
        super(SCSSFileStorage, self).__init__(location, base_url,
                                                *args, **kwargs)



from django.conf.global_settings import *
import os


DEBUG = True

STATIC_ROOT = MEDIA_ROOT = os.path.join(os.path.dirname(__file__), 'static')
STATIC_URL = MEDIA_URL = "/static/"

STATICFILES_DIRS = (
    os.path.join(os.path.dirname(__file__), 'staticfiles_dir'),
    ("prefix", os.path.join(os.path.dirname(__file__), 'staticfiles_dir_with_prefix')),
)


INSTALLED_APPS = (
    "django_scss",
)
SCSS_MTIME_DELAY = 2
SCSS_OUTPUT_DIR = "SCSS_CACHE"

LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'handlers': {
        'console':{
            'level':'DEBUG',
            'class':'logging.StreamHandler',
        },
    },
    'loggers': {
        'django_scss': {
            'handlers': ['console'],
            'level': 'DEBUG',
        },
    }
}



from ..cache import get_cache_key, get_hexdigest, get_hashed_mtime
from ..settings import SCSS_EXECUTABLE, SCSS_USE_CACHE,\
    SCSS_CACHE_TIMEOUT, SCSS_ROOT, SCSS_OUTPUT_DIR, SCSS_DEVMODE,\
    SCSS_DEVMODE_WATCH_DIRS
from ..utils import compile_scss, STATIC_ROOT
from django.conf import settings
from django.core.cache import cache
from django.contrib.staticfiles import finders
from django.template.base import Library, Node, TemplateSyntaxError
import logging
import shlex
import subprocess
import os
import sys


logger = logging.getLogger("django_scss")


register = Library()


class InlineSCSSNode(Node):

    def __init__(self, nodelist):
        self.nodelist = nodelist

    def compile(self, source):


        args = shlex.split("%s -s --scss -C" % SCSS_EXECUTABLE)

        p = subprocess.Popen(args, stdin=subprocess.PIPE,
                             stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        out, errors = p.communicate(source)
        if out:
            return out.decode(settings.FILE_CHARSET)
        elif errors:
            return errors.decode(settings.FILE_CHARSET)

        return u""

    def render(self, context):
        output = self.nodelist.render(context)

        if SCSS_USE_CACHE:
            cache_key = get_cache_key(get_hexdigest(output))
            cached = cache.get(cache_key, None)
            if cached is not None:
                return cached
            output = self.compile(output)
            cache.set(cache_key, output, SCSS_CACHE_TIMEOUT)
            return output
        else:
            return self.compile(output)


@register.tag(name="inlinescss")
def do_inlinescss(parser, token):
    nodelist = parser.parse(("endinlinescss",))
    parser.delete_first_token()
    return InlineSCSSNode(nodelist)


def scss_paths(path):

    full_path = os.path.join(STATIC_ROOT, path)

    if settings.DEBUG and not os.path.exists(full_path):
        # while developing it is more confortable
        # searching for the scss files rather then
        # doing collectstatics all the time
        full_path = finders.find(path)

        if full_path is None:
            raise TemplateSyntaxError("Can't find staticfile named: {}".format(path))

    file_name = os.path.split(path)[-1]
    output_dir = os.path.join(SCSS_ROOT, SCSS_OUTPUT_DIR, os.path.dirname(path))

    return full_path, file_name, output_dir


@register.simple_tag
def scss(path):

    logger.info("processing file %s" % path)

    full_path, file_name, output_dir = scss_paths(path)
    base_file_name = os.path.splitext(file_name)[0]

    if SCSS_DEVMODE and any(map(lambda watched_dir: full_path.startswith(watched_dir), SCSS_DEVMODE_WATCH_DIRS)):
        return os.path.join(os.path.dirname(path), "%s.css" % base_file_name)

    hashed_mtime = get_hashed_mtime(full_path)
    output_file = "%s-%s.css" % (base_file_name, hashed_mtime)
    output_path = os.path.join(output_dir, output_file)

    encoded_full_path = full_path
    if isinstance(full_path, unicode):
        filesystem_encoding = sys.getfilesystemencoding() or sys.getdefaultencoding()
        encoded_full_path = full_path.encode(filesystem_encoding)

    if not os.path.exists(output_path):
        if not compile_scss(encoded_full_path, output_path, path):
            return path

        # Remove old files
        compiled_filename = os.path.split(output_path)[-1]
        for filename in os.listdir(output_dir):
            if filename.startswith(base_file_name) and filename != compiled_filename:
                os.remove(os.path.join(output_dir, filename))

    return os.path.join(SCSS_OUTPUT_DIR, os.path.dirname(path), output_file)



from django_scss.settings import SCSS_MTIME_DELAY
from django.core.cache import cache
from django.utils.encoding import smart_str
from django.utils.hashcompat import md5_constructor
import os.path
import socket


def get_hexdigest(plaintext, length=None):
    digest = md5_constructor(smart_str(plaintext)).hexdigest()
    if length:
        return digest[:length]
    return digest


def get_cache_key(key):
    return ("django_scss.%s.%s" % (socket.gethostname(), key))


def get_mtime_cachekey(filename):
    return get_cache_key("mtime.%s" % get_hexdigest(filename))


def get_mtime(filename):
    if SCSS_MTIME_DELAY:
        key = get_mtime_cachekey(filename)
        mtime = cache.get(key)
        if mtime is None:
            mtime = os.path.getmtime(filename)
            cache.set(key, mtime, SCSS_MTIME_DELAY)
        return mtime
    return os.path.getmtime(filename)


def get_hashed_mtime(filename, length=12):
    try:
        filename = os.path.realpath(filename)
        mtime = str(int(get_mtime(filename)))
    except OSError:
        return None
    return get_hexdigest(mtime, length)



from django_scss.settings import SCSS_EXECUTABLE, SCSS_USE_COMPASS, SCSS_ROOT,\
    SCSS_OUTPUT_DIR
from django.conf import settings
import logging
import os
import posixpath
import re
import subprocess


logger = logging.getLogger("django_scss")


STATIC_ROOT = getattr(settings, "STATIC_ROOT", getattr(settings, "MEDIA_ROOT"))
STATIC_URL = getattr(settings, "STATIC_URL", getattr(settings, "MEDIA_URL"))


class URLConverter(object):

    URL_PATTERN = re.compile(r'url\(([^\)]+)\)')

    def __init__(self, content, source_path):
        self.content = content
        self.source_dir = os.path.dirname(source_path)

    def convert_url(self, matchobj):
        url = matchobj.group(1)
        url = url.strip(' \'"')
        if url.startswith(('http://', 'https://', '/', 'data:')):
            return "url('%s')" % url
        full_url = posixpath.normpath("/".join([self.source_dir, url]))
        return "url('%s')" % full_url

    def convert(self):
        return self.URL_PATTERN.sub(self.convert_url, self.content)


def compile_scss(input, output, scss_path):

    scss_root = os.path.join(SCSS_ROOT, SCSS_OUTPUT_DIR)
    if not os.path.exists(scss_root):
        os.makedirs(scss_root)

    args = [SCSS_EXECUTABLE, "-C", input]
    if SCSS_USE_COMPASS:
        args.insert(1, "--compass")

    p = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=STATIC_ROOT)
    out, errors = p.communicate()

    if errors:
        logger.error(errors)
        return False

    output_directory = os.path.dirname(output)
    if not os.path.exists(output_directory):
        os.makedirs(output_directory)
    compiled_css = URLConverter(
        out.decode(settings.FILE_CHARSET),
        os.path.join(STATIC_URL, scss_path)
    ).convert()
    compiled_file = open(output, "w+")
    compiled_file.write(compiled_css.encode(settings.FILE_CHARSET))
    compiled_file.close()

    return True




from setuptools import setup, find_packages
import os


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


README = read('README.rst')
CHANGES = read('CHANGES.rst')


setup(
    name = "django-scss",
    packages = find_packages(),
    version = "0.6",
    author = "Andrey Fedoseev",
    author_email = "andrey.fedoseev@gmail.com",
    url = "https://github.com/andreyfedoseev/django-scss",
    description = "Django template tags to compile SCSS into CSS",
    long_description = "\n\n".join([README, CHANGES]),
    classifiers = [
        'Development Status :: 4 - Beta',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Internet :: WWW/HTTP',
    ],
    keywords = ["scss", "css"],
)



# coding: utf-8
from unittest import main, TestCase
from django.http import HttpRequest
from django.template.base import Template
from django.template.context import RequestContext
import os
import re
import time
import shutil


os.environ["DJANGO_SETTINGS_MODULE"] = "django_scss.tests.django_settings"


class SCSSTestCase(TestCase):

    def setUp(self):
        from django.conf import settings as django_settings
        self.django_settings = django_settings

        output_dir = os.path.join(self.django_settings.STATIC_ROOT,
                                  self.django_settings.SCSS_OUTPUT_DIR)

        # Remove the output directory if it exists to start from scratch
        if os.path.exists(output_dir):
            shutil.rmtree(output_dir)

    def _get_request_context(self):
        return RequestContext(HttpRequest())

    def test_inline_scss(self):
        template = Template("""
        {% load scss %}
        {% inlinescss %}
            $the-border: 1px;
            #bordered {
                border: $the-border * 2;
            }
        {% endinlinescss %}
        """)
        rendered = """#bordered {
  border: 2px; }"""
        self.assertEqual(template.render(self._get_request_context()).strip(), rendered)

    def test_external_scss(self):

        template = Template("""
        {% load scss %}
        {% scss "styles/test.scss" %}
        """)
        compiled_filename_re = re.compile(r"SCSS_CACHE/styles/test-[a-f0-9]{12}.css")
        compiled_filename = template.render(self._get_request_context()).strip()
        self.assertTrue(bool(compiled_filename_re.match(compiled_filename)))

        compiled_path = os.path.join(self.django_settings.STATIC_ROOT, compiled_filename)
        compiled_content = open(compiled_path).read().strip()
        compiled = """#header h1 {
  background-image: url('/static/images/header.png'); }"""
        self.assertEquals(compiled_content, compiled)

        # Change the modification time
        source_path = os.path.join(self.django_settings.STATIC_ROOT, "styles/test.scss")
        os.utime(source_path, None)

        # The modification time is cached so the compiled file is not updated
        compiled_filename_2 = template.render(self._get_request_context()).strip()
        self.assertTrue(bool(compiled_filename_re.match(compiled_filename_2)))
        self.assertEquals(compiled_filename, compiled_filename_2)

        # Wait to invalidate the cached modification time
        time.sleep(self.django_settings.SCSS_MTIME_DELAY)

        # Now the file is re-compiled
        compiled_filename_3 = template.render(self._get_request_context()).strip()
        self.assertTrue(bool(compiled_filename_re.match(compiled_filename_3)))
        self.assertNotEquals(compiled_filename, compiled_filename_3)

        # Check that we have only one compiled file, old files should be removed

        compiled_file_dir = os.path.dirname(os.path.join(self.django_settings.STATIC_ROOT,
                                                         compiled_filename_3))
        self.assertEquals(len(os.listdir(compiled_file_dir)), 1)

    def test_lookup_in_staticfiles_dirs(self):

        template = Template("""
        {% load scss %}
        {% scss "another_test.scss" %}
        """)
        compiled_filename_re = re.compile(r"SCSS_CACHE/another_test-[a-f0-9]{12}.css")
        compiled_filename = template.render(self._get_request_context()).strip()
        self.assertTrue(bool(compiled_filename_re.match(compiled_filename)))

        compiled_path = os.path.join(self.django_settings.STATIC_ROOT, compiled_filename)
        compiled_content = open(compiled_path).read().strip()
        compiled = """#header-from-staticfiles-dir h1 {
  color: red; }"""
        self.assertEquals(compiled_content, compiled)

        template = Template("""
        {% load scss %}
        {% scss "prefix/another_test.scss" %}
        """)
        compiled_filename_re = re.compile(r"SCSS_CACHE/prefix/another_test-[a-f0-9]{12}.css")
        compiled_filename = template.render(self._get_request_context()).strip()
        self.assertTrue(bool(compiled_filename_re.match(compiled_filename)))

        compiled_path = os.path.join(self.django_settings.STATIC_ROOT, compiled_filename)
        compiled_content = open(compiled_path).read().strip()
        compiled = """#header-from-staticfiles-dir-with-prefix h1 {
  color: red; }"""
        self.assertEquals(compiled_content, compiled)

    def test_non_ascii_content(self):

        template = Template("""
        {% load scss %}
        {% scss "styles/non-ascii.scss" %}
        """)
        compiled_filename = template.render(self._get_request_context()).strip()
        compiled_path = os.path.join(self.django_settings.STATIC_ROOT, compiled_filename)
        compiled_content = open(compiled_path).read().strip()
        compiled = """.external_link:first-child:before {
  content: "Zobacz także:";
  background: url('/static/styles/картинка.png'); }"""
        self.assertEquals(compiled_content, compiled)


if __name__ == '__main__':
    main()



from django_scss.settings import SCSS_DEVMODE


if SCSS_DEVMODE:
    # Run the devmode daemon if it's enabled.
    # We start it here because this file is auto imported by Django when
    # devserver is started.
    from django_scss.devmode import start_daemon
    start_daemon()
